/*
  Copyright (C) 2012  Torsten Brischalle

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/
*/

#ifndef LUATHREAD_H
#define LUATHREAD_H

//-----------------------------------------------------------------------------

#include <QThread>
#include <QSemaphore>
#include <QVariant>
#include <QVariantList>
#include <QMutex>

#include <stdexcept>
#include <string>
#include <deque>

#include "lua.h"

#include "InterfaceGUI.h"
#include "InterfaceBreakpoints.h"
#include "callstack.h"
#include "variablelist.h"

//-----------------------------------------------------------------------------

class LuaThread
: public QThread
, public InterfaceGUI
, public InterfaceBreakpoints
{
  Q_OBJECT
public:
  explicit LuaThread(QObject *tab, QObject *gfx, QObject *dia, QObject *vecgfx);
  virtual ~LuaThread();

  void Execute(const QString & luaCode);

  void PutToStdInput(const QString & str);

  bool IsRunning() const { return isRunning(); }

  void ResetLua();

  void Debug(bool debugOn);
  void DebuggerStepIn();
  void DebuggerStepOut();
  void DebuggerStepOver();
  void DebuggerContinue();
  void DebuggerHalt();

  CallStack DebuggerCallStack();
  VariableList DebuggerGetLocalVariables();
  VariableList DebuggerGetGlobalVariables();
  VariableList DebuggerGetLocalVariablesFromPath(const QVariantList & path);
  VariableList DebuggerGetGlobalVariablesFromPath(const QVariantList & path);

  virtual void ClearBreakpoints();
  virtual void ClearBreakpoint(int line);
  virtual void SetBreakpoint(int line);
  virtual bool IsBreakpoint(int line);
  virtual bool DebugEnabled();
  virtual int  CurrentLine();

  virtual void TableClear();
  virtual void TableSetSize(int rows, int columns);
  virtual void TableSetValue(int row, int column, const QString & value);
  virtual void TableSetValuesInRow(int row, const QList <QString> & valueList);
  virtual void TableSetValuesInColumn(int column, const QList <QString> & valueList);
  virtual void TableSaveCsv(const QString & fileName);

  virtual void GfxClear();
  virtual void GfxSetSize(int x, int y);
  virtual void GfxDrawPoint(int x, int y);
  virtual void GfxDrawLine(int x1, int y1, int x2, int y2);
  virtual void GfxDrawCircle(int x, int y, int r);
  virtual void GfxDrawEllipse(int x, int y, int rx, int ry);
  virtual void GfxDrawRect(int x, int y, int w, int h);
  virtual void GfxDrawPolygon(const QPolygon & polygon);
  virtual void GfxFillCircle(int x, int y, int r);
  virtual void GfxFillEllipse(int x, int y, int rx, int ry);
  virtual void GfxFillRect(int x, int y, int w, int h);
  virtual void GfxFillPolygon(const QPolygon & polygon);
  virtual void GfxDrawText(int x, int y, const QString & text);
  virtual void GfxSetColor(int red, int green, int blue);
  virtual void GfxSetColor(const QString & name);
  virtual void GfxSetFillColor(int red, int green, int blue);
  virtual void GfxSetFillColor(const QString & name);
  virtual void GfxBeginUpdate();
  virtual void GfxEndUpdate();
  virtual void GfxSaveImg(const QString & fileName);

  virtual void DiaDiagramNew(const QString & diagramName);
  virtual void DiaDiagramDelete(const QString & diagramName);
  virtual void DiaDiagramSaveImg(const QString & diagramName, const QString & fileName);
  virtual void DiaDiagramSaveCsv(const QString & diagramName, const QString & fileName);
  virtual void DiaDiagramDeleteAll();
  virtual void DiaCurveNew(const QString & diagramName, const QString & curveName, int yAxis);
  virtual void DiaCurveDelete(const QString & diagramName, const QString & curveName);
  virtual void DiaCurveDeleteAll(const QString & diagramName);
  virtual void DiaCurveClearData(const QString & diagramName, const QString & curveName);
  virtual void DiaCurveSetData(const QString & diagramName, const QString &curveName, const QVector <double> & x, const QVector <double> & y);
  virtual void DiaCurveAppendData(const QString & diagramName, const QString &curveName, double x, double y);

  virtual void VecGfxClear();
  virtual void VecGfxSetColor(int red, int green, int blue);
  virtual void VecGfxSetColor(const QString & name);
  virtual void VecGfxSetFillColor(int red, int green, int blue);
  virtual void VecGfxSetFillColor(const QString & name);
  virtual void VecGfxDrawLine(qreal x1, qreal y1, qreal x2, qreal y2);
  virtual void VecGfxDrawCircle(qreal x, qreal y, qreal r);
  virtual void VecGfxDrawEllipse(qreal x, qreal y, qreal rx, qreal ry);
  virtual void VecGfxDrawRect(qreal x, qreal y, qreal w, qreal h);
  virtual void VecGfxDrawPolygon(const QPolygonF & polygon);
  virtual void VecGfxFillCircle(qreal x, qreal y, qreal r);
  virtual void VecGfxFillEllipse(qreal x, qreal y, qreal rx, qreal ry);
  virtual void VecGfxFillRect(qreal x, qreal y, qreal w, qreal h);
  virtual void VecGfxFillPolygon(const QPolygonF & polygon);
  virtual void VecGfxDrawText(qreal x, qreal y, const QString & text);
  virtual void VecGfxSaveImg(const QString & fileName, int width, int height);
  virtual void VecGfxSaveSvg(const QString & fileName);

  virtual bool ConsoleReadNumber(double *number);
  virtual bool ConsoleTestEof();
  virtual bool ConsoleReadLine(char *str, int maxNum);
  virtual int  ConsoleReadChars(char *str, int count);

signals:
  void LuaStarted();
  void LuaStopped();
  void LuaHalted();
  void LuaContinued();

  void SignalTableClear();
  void SignalTableSetSize(int rows, int columns);
  void SignalTableSetValue(int row, int column, const QString & value);
  void SignalTableSetValuesInRow(int row, const QList <QString> & valueList);
  void SignalTableSetValuesInColumn(int column, const QList <QString> & valueList);
  void SignalTableSaveCsv(const QString & fileName);

  void SignalGfxClear();
  void SignalGfxSetSize(int x, int y);
  void SignalGfxDrawPoint(int x, int y);
  void SignalGfxDrawLine(int x1, int y1, int x2, int y2);
  void SignalGfxDrawCircle(int x, int y, int r);
  void SignalGfxDrawEllipse(int x, int y, int rx, int ry);
  void SignalGfxDrawRect(int x, int y, int w, int h);
  void SignalGfxDrawPolygon(const QPolygon & polygon);
  void SignalGfxFillCircle(int x, int y, int r);
  void SignalGfxFillEllipse(int x, int y, int rx, int ry);
  void SignalGfxFillRect(int x, int y, int w, int h);
  void SignalGfxFillPolygon(const QPolygon & polygon);
  void SignalGfxDrawText(int x, int y, const QString & text);
  void SignalGfxSetColor(int red, int green, int blue);
  void SignalGfxSetColor(const QString & name);
  void SignalGfxSetFillColor(int red, int green, int blue);
  void SignalGfxSetFillColor(const QString & name);
  void SignalGfxBeginUpdate();
  void SignalGfxEndUpdate();
  void SignalGfxSaveImg(const QString & fileName);

  void SignalDiaDiagramNew(const QString & diagramName);
  void SignalDiaDiagramDelete(const QString & diagramName);
  void SignalDiaDiagramSaveImg(const QString & diagramName, const QString & fileName);
  void SignalDiaDiagramSaveCsv(const QString & diagramName, const QString & fileName);
  void SignalDiaDiagramDeleteAll();
  void SignalDiaCurveNew(const QString & diagramName, const QString & curveName, int yAxis);
  void SignalDiaCurveDelete(const QString & diagramName, const QString & curveName);
  void SignalDiaCurveDeleteAll(const QString & diagramName);
  void SignalDiaCurveClearData(const QString & diagramName, const QString & curveName);
  void SignalDiaCurveSetData(const QString & diagramName, const QString &curveName, const QVector <double> & x, const QVector <double> & y);
  void SignalDiaCurveAppendData(const QString & diagramName, const QString &curveName, double x, double y);

  void SignalVecGfxClear();
  void SignalVecGfxSetColor(int red, int green, int blue);
  void SignalVecGfxSetColor(const QString & name);
  void SignalVecGfxSetFillColor(int red, int green, int blue);
  void SignalVecGfxSetFillColor(const QString & name);
  void SignalVecGfxDrawLine(qreal x1, qreal y1, qreal x2, qreal y2);
  void SignalVecGfxDrawCircle(qreal x, qreal y, qreal r);
  void SignalVecGfxDrawEllipse(qreal x, qreal y, qreal rx, qreal ry);
  void SignalVecGfxDrawRect(qreal x, qreal y, qreal w, qreal h);
  void SignalVecGfxDrawPolygon(const QPolygonF & polygon);
  void SignalVecGfxFillCircle(qreal x, qreal y, qreal r);
  void SignalVecGfxFillEllipse(qreal x, qreal y, qreal rx, qreal ry);
  void SignalVecGfxFillRect(qreal x, qreal y, qreal w, qreal h);
  void SignalVecGfxFillPolygon(const QPolygonF & polygon);
  void SignalVecGfxDrawText(qreal x, qreal y, const QString & text);
  void SignalVecGfxSaveImg(const QString & fileName, int width, int height);
  void SignalVecGfxSaveSvg(const QString & fileName);

protected:
  void run();

private:

  enum { DEBUGGER_MAX_BREAKPOINTS = 64 };

  volatile int  m_debuggerBreakpoints[DEBUGGER_MAX_BREAKPOINTS];

  enum DebuggerCmd
  {
    DEBUGGER_CMD_NONE = 0,
    DEBUGGER_CMD_HALT,
    DEBUGGER_CMD_CONTINUE,
    DEBUGGER_CMD_STEP_IN,
    DEBUGGER_CMD_STEP_OUT,
    DEBUGGER_CMD_STEP_OVER
  };

  QString       m_luaCode;
  lua_State     *m_lua;

  volatile bool m_luaRunning;
  volatile int  m_currentLine;

  volatile bool         m_debuggerHalted;
  volatile DebuggerCmd  m_debuggerCmd;
  QSemaphore            m_debuggerCmdSemaphore;
  volatile int          m_stackDepth;

  std::deque <char>     m_stdInput;
  QMutex                m_stdInputMutex;

  static void debugger_lua_hook(lua_State *lua, lua_Debug *ar);

  void DebuggerLuaHook(lua_State *lua, lua_Debug *ar);

  void PrintLocalVariables();
  void PrintGlobalVariables();

  int GetLuaRuntimeStackDepth();

  void ClearStdInput();
};

//-----------------------------------------------------------------------------

#endif // LUATHREAD_H
